#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/xpm.h>
#include <sys/time.h>
#include <signal.h>
#include <limits.h>
#include <getopt.h>

/**********************************************************/
/* misc definitions */

const char *ProgramName="xmuwin";
const char *Version="version 1.2.1";
const char *Copyright="graphics by Mominoha. Copyright (C) 1998,1999 Mominoha.\nMS-Windows version programmed by Narui. Copyright (C) 1998,1999 Narui.\nX11 version programmed by Nakatsuka. Copyright (C) 1999 Nakatsuka.";

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef BOOL
#define BOOL int
#endif

#ifdef ENABLEALLOCA
#define MALLOC alloca
#define FREE(u)
#else
#define MALLOC malloc
#define FREE(u) free(u)
#endif

typedef void (*sighandler_t)(int);

/**********************************************************/
/* pixmap */

#include "mu0.xpm"
#include "mu1.xpm"
#include "mu2.xpm"
#include "mu3.xpm"
#include "mu4.xpm"
#include "mu5.xpm"
#include "mu6.xpm"
#include "mu7.xpm"

/**********************************************************/
/* options */

int opt_framerate=3; /* framerate of frame */
int opt_probability=3; /* appearance probability */
int opt_poscheck=FALSE; /* position check */
int opt_targetoverride=FALSE; /* target override window */
int opt_nobite=FALSE; /* no bite */

/**********************************************************/
/* constant variables of frames */

#define IMG_NUM 8
#define IMG_SIZEX 105
#define IMG_SIZEY 193
#define IMG_CENTERX 75
#define IMG_CENTERY 60

#define MINWINSIZEX 30
#define MINWINSIZEY 133

const int frame_num=85;
const int frame_ptn[]={
  7,6,5,
  0,0,0,0,1,1,1,1,
  0,0,0,0,0,0,0,0,
  0,0,0,0,1,1,1,1,0,0,0,0,1,1,1,1,
  0,0,0,1,1,1,0,0,0,1,1,1,
  0,0,0,0,0,0,0,0,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,
  3,4,4,4,3,
  2,2,2,2,2,2,2,2,
  5,6,7,
};

/**********************************************************/
/* global variables */

Pixmap img_pixmap[IMG_NUM],img_mask[IMG_NUM];
char *my_displayname=NULL;
Display *my_display;
Window my_window;
int my_screennum;
GC my_gc;

int anm_inanim=FALSE;
int anm_frameno=0;
int anm_timercount=0;
int anm_ignitetime;
Window attach_win;

BOOL LoadPixmap(void)
{
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm0,&img_pixmap[0],&img_mask[0],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm1,&img_pixmap[1],&img_mask[1],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm2,&img_pixmap[2],&img_mask[2],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm3,&img_pixmap[3],&img_mask[3],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm4,&img_pixmap[4],&img_mask[4],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm5,&img_pixmap[5],&img_mask[5],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm6,&img_pixmap[6],&img_mask[6],0)!=XpmSuccess)
    return 0;
  if(XpmCreatePixmapFromData(my_display,my_window,muxpm7,&img_pixmap[7],&img_mask[7],0)!=XpmSuccess)
    return 0;
  return 1;
}

void ReleasePixmap(void)
{
  int i;
  for(i=0;i<IMG_NUM;i++) {
    XFreePixmap(my_display,img_pixmap[i]);
    XFreePixmap(my_display,img_mask[i]);
  }
}

BOOL SortWindow(void) {
  XWindowChanges changes;
  changes.sibling=attach_win;
  changes.stack_mode=Above;
  XConfigureWindow(my_display,my_window,CWSibling|CWStackMode,&changes);
}

BOOL IsValidWindow(Window w)
{
  XWindowAttributes attr;
  Window root,parent,*children;
  int nchildren;
  int i;

  if(BadWindow==XGetWindowAttributes(my_display,w,&attr))
    return FALSE;

  if(!(attr.map_state == 2   /* visible */
       && attr.width  > MINWINSIZEX
       && attr.height > MINWINSIZEY /* window size check */
       && ( !opt_poscheck || 
	    (attr.x > IMG_CENTERX
	     && attr.y > IMG_CENTERY
	     && attr.x < DisplayWidth(my_display,my_screennum)-MINWINSIZEX
	     && attr.y < DisplayHeight(my_display,my_screennum)-MINWINSIZEY))
       /* window position check*/
       ))
    return FALSE;

  /* get window list */
  if(!XQueryTree(my_display,w,
		 &root,
		 &parent,
		 &children,
		 &nchildren))
    return FALSE;

  if(nchildren==0)
    if(attr.override_redirect)
      return opt_targetoverride; /* if opt is TRUE target override window */
    else 
      return TRUE;

  for(i=0;i<nchildren;i++) {
    if(BadWindow==XGetWindowAttributes(my_display,children[i],&attr))
      return FALSE;
    if(!attr.override_redirect) {
      XFree(children);
      return TRUE;
    }
  }
  XFree(children);
  return opt_targetoverride; /* if opt is TRUE target override window */
}

void SetBiteShape(void)
{
  Pixmap bitmap;
  GC gc;
  XWindowAttributes attr;
  XPoint point[3];
  int width,height;

  XGetWindowAttributes(my_display,attach_win,&attr);
  width=(attr.width > DisplayWidth(my_display,my_screennum)) ? attr.width : DisplayWidth(my_display,my_screennum);
  height=(attr.height > DisplayHeight(my_display,my_screennum)) ? attr.height : DisplayHeight(my_display,my_screennum);

  bitmap=XCreatePixmap(my_display,my_window,
		       width+attr.border_width*2,
		       height+attr.border_width*2,1);
  gc=XCreateGC(my_display,bitmap,0,NULL);

  XSetForeground(my_display,gc,1);
  XFillRectangle(my_display,bitmap,gc,0,0,
		 width+attr.border_width*2,
		 height+attr.border_width*2);
  XSetForeground(my_display,gc,0);
  point[0].x=0;point[0].y=0;
  point[1].x=0;point[1].y=4;
  point[2].x=4;point[2].y=0;
  XFillPolygon(my_display,bitmap,gc,point,3,Convex,CoordModeOrigin);
  XShapeCombineMask(my_display,attach_win,ShapeBounding,
		    -attr.border_width,
		    -attr.border_width,bitmap,ShapeSet);

  XSetForeground(my_display,gc,0);
  XFillRectangle(my_display,bitmap,gc,0,0,
		 width+attr.border_width*2,
		 height+attr.border_width*2);

  XSetForeground(my_display,gc,1);
  XFillRectangle(my_display,bitmap,gc,0,0,
		 width+attr.border_width,
		 height+attr.border_width);

  XSetForeground(my_display,gc,0);
  point[0].x=0;point[0].y=0;
  point[1].x=0;point[1].y=5;
  point[2].x=5;point[2].y=0;
  XFillPolygon(my_display,bitmap,gc,point,3,Convex,CoordModeOrigin);
  XShapeCombineMask(my_display,attach_win,ShapeClip,
		    -attr.border_width,
		    -attr.border_width,bitmap,ShapeSet);

  XFreeGC(my_display,gc);
  XFreePixmap(my_display,bitmap);

}

void Animate(int no)
{
  XWindowAttributes attr;

  SortWindow();

  /* window pos check */
  if(!IsValidWindow(attach_win)) {
    anm_inanim=FALSE;
    XUnmapWindow(my_display,my_window);
    XFlush(my_display);
    return;
  }

  /* move window */
  XGetWindowAttributes(my_display,attach_win,&attr);
  XMoveWindow(my_display,my_window,
	      attr.x-IMG_CENTERX,
	      attr.y-IMG_CENTERY);

  /* draw pixmap */
  XShapeCombineMask(my_display,my_window,
		    ShapeBounding,0,0,img_mask[frame_ptn[no]],ShapeSet);
  XCopyArea(my_display,
	    img_pixmap[frame_ptn[no]],my_window,my_gc,0,0,IMG_SIZEX,IMG_SIZEY,0,0);
  XFlush(my_display);
}

BOOL GetAttachWindow(Window *w)
{
  Window root,parent,*children;
  Window *cand;
  int nchildren;
  int i;
  int winnum=0;

  /* get window list */
  XQueryTree(my_display,RootWindow(my_display,0),
	     &root,
	     &parent,
	     &children,
	     &nchildren);

  /* get valid window */
  cand=(Window *)MALLOC(sizeof(Window)*nchildren);
  if(!cand) {
    fprintf(stderr,"cannot allocate memory.\n");
    signal(SIGALRM,SIG_DFL);
    
    XSetErrorHandler(NULL);
    
    XFreeGC(my_display,my_gc);
    
    XUnmapWindow(my_display,my_window);
    ReleasePixmap();
    XDestroyWindow(my_display,my_window);
    XCloseDisplay(my_display);
    exit(0);
  }

  for(i=0;i<nchildren;i++) {
    if(IsValidWindow(children[i])) {
      cand[winnum]=children[i];
      winnum++;
    }
  }

  /* select window */
  if(winnum==0) {
    XFree(children);
    FREE(cand);
    return FALSE;
  }
  i=(int) ((double)winnum*rand()/(RAND_MAX+1.0));
  *w=cand[i];
  XFree(children);
  FREE(cand);
  return TRUE;
}

void OnSigAlarm(void)
{
  signal(SIGALRM,SIG_DFL);
  if(anm_inanim) {
    anm_frameno++;
    if(anm_frameno == frame_num) {
      if(!opt_nobite) SetBiteShape();
      XUnmapWindow(my_display,my_window);
      XFlush(my_display);
      anm_inanim=FALSE;
    } else
      Animate(anm_frameno);
  } else {
    if(anm_ignitetime<anm_timercount) {
      if(GetAttachWindow(&attach_win)) {
	XMapWindow(my_display,my_window );
	anm_frameno=0;
	anm_inanim=TRUE;
	anm_timercount=0;
	anm_ignitetime=(80+opt_probability*140)/opt_framerate;
	Animate(anm_frameno);
      } else {
	anm_timercount=0;
	anm_ignitetime=(80+opt_probability*140)/opt_framerate;
      }
    } else
      anm_timercount++;
    /*printf("%d %d\n",anm_ignitetime,anm_timercount);*/
  }
  signal(SIGALRM,(sighandler_t) OnSigAlarm);
}

static int (*NextErrorHandler)(Display *d, XErrorEvent *ev);

static int OnXError(Display *d, XErrorEvent *ev)
{
  if (ev->error_code == BadWindow) {
    anm_inanim=FALSE;
    XUnmapWindow(my_display,my_window);
    return 0;
  }
  else    
    return NextErrorHandler(d,ev);
}

void ShowHelp(void)
{
  printf("\n%s %s\n",ProgramName,Version);
  printf("%s\n\n",Copyright);
  printf("-h , --help              print this text\n");
  printf("-d , --display <host:n>  specify display\n");
  printf("-f , --framerate <n>     specify framerate. n is between 1 and 5.\n");
  printf("-p , --probability <n>   specify appearance probability. n is between 1 and 5.\n");
  printf("-b , --nobite            no biting.\n");
  printf("-c , --poscheck          position check.\n");
  printf("-o , --override          enable targetiing override_redirect window.\n");
  printf("\n");
}

BOOL ParseArgument(int argc,char *argv[])
{
  int opt;
  int option_index;
  struct option long_options[] = {
    {"display", 1, 0, 'd'},
    {"framerate", 1, 0, 'i'},
    {"probability", 1, 0, 'p'},
    {"nobite", 0, 0, 'b'},
    {"poscheck", 0, 0, 'c'},
    {"override", 0, 0, 'o'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  while((opt=getopt_long(argc,argv,"hd:f:p:bco",long_options,&option_index)) != -1) {
    switch(opt) {
    case 'd':
      if(optarg) {
	my_displayname=malloc(strlen(optarg));
	strcpy(my_displayname,optarg);
      } else {
	printf("display is not specified.\n");
	return FALSE;
      }
      break;
    case 'f':
      if(optarg) {
	if(*optarg <= '0' || *optarg >= '6') {
	  printf("invalid framerate\n");
	  return FALSE;
	}
	opt_framerate=atoi(optarg);
      } else{
	printf("no framerate argument\n");
	return FALSE;
      }
      break;
    case 'p':
      if(optarg) {
	if(*optarg <= '0' || *optarg >= '6') {
	  printf("invalid probability\n");
	  return FALSE;
	}
	opt_probability=atoi(optarg);
      } else{
	printf("no probability argument\n");
	return FALSE;
      }
      break;
    case 'c':
      opt_poscheck=TRUE;
      break;
    case 'b':
      opt_nobite=TRUE;
      break;
    case 'o':
      opt_targetoverride=TRUE;
      break;
    case '?':
      /*printf("unknown option : %c\n",optopt);*/
      return FALSE;
    case ':':
      printf("option needs a value\n");
      return FALSE;
    case 'h':
      ShowHelp();
      return FALSE;
    default:
      printf("getopt returned character code 0%o \n", opt);
      return FALSE;
    }
  }

#ifdef DEBUG
  printf("framerate:%d\n",opt_framerate);
  printf("probability:%d\n",opt_probability);
  printf("nobite:%d\n",opt_nobite);
  printf("poscheck:%d\n",opt_poscheck);
  printf("targetoverride:%d\n",opt_targetoverride);
#endif
  return TRUE;
}

int main(int argc,char *argv[])
{
  XSetWindowAttributes attr;
  XEvent event;
  int loop=1;
  struct itimerval itime;

  srand((unsigned)getpid());

  if(!ParseArgument(argc,argv)) return 0;

  if(!(my_display = XOpenDisplay(my_displayname))) {
    fprintf(stderr,"cannot open display.\n");
    return 0;
  }
  my_screennum=DefaultScreen(my_display);
  attr.save_under=True;
  attr.override_redirect=True;
  my_window=XCreateWindow(my_display,RootWindow(my_display,my_screennum),
		  0,0,
		  IMG_SIZEX,IMG_SIZEY,
		  0,DefaultDepth(my_display,my_screennum),
		  InputOutput,
		  CopyFromParent,
		  CWSaveUnder | CWOverrideRedirect,
		  &attr);

  if(!LoadPixmap()) {
    fprintf(stderr,"can't create pixmaps. probably palette runs short.\n");
    XDestroyWindow(my_display,my_window);
    XCloseDisplay(my_display);
    return 0;
  }

  my_gc=XCreateGC(my_display,my_window,0,0);
  XSetGraphicsExposures(my_display,my_gc,False);

  NextErrorHandler=XSetErrorHandler(OnXError);

  /* event */
  XSelectInput(my_display,my_window,ButtonPressMask);

  /* timer handler */
  signal(SIGALRM,(sighandler_t) OnSigAlarm);

  timerclear(&itime.it_interval);
  timerclear(&itime.it_value);
  itime.it_interval.tv_usec=opt_framerate*20000L;
  itime.it_value.tv_usec=opt_framerate*20000L;
  setitimer(ITIMER_REAL,&itime,0);

  while(loop) {

    /* animation process */
    
    /* event process */
    XNextEvent(my_display,&event);
    
    switch(event.type) {
    case ButtonPress:
      switch(event.xbutton.button) {
      case 1:
	/* window unmap not to be obstacle */
	anm_inanim=FALSE;
	XUnmapWindow(my_display,my_window);
	break;
      case 3:
	/* exit program */
	loop=0;
	break;
      default:
	break;
      }
      break;
    case Expose:
      if(anm_inanim) Animate(anm_frameno);
      break;
    default:
      break;
    }
  }
  signal(SIGALRM,SIG_DFL);

  XSetErrorHandler(NULL);

  XFreeGC(my_display,my_gc);

  XUnmapWindow(my_display,my_window);
  ReleasePixmap();
  XDestroyWindow(my_display,my_window);
  XCloseDisplay(my_display);
  if(my_displayname!=NULL) free(my_displayname);
  return 0;
}



